# -*- coding: utf-8 -*-

"""
/***************************************************************************
 OpenLidarTools
                                 A QGIS QGISplugin
 Open LiDAR Toolbox
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2021-03-10
        copyright            : (C) 2021 by Benjamin Štular, Edisa Lozić, Stefan Eichert
        email                : stefaneichert@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = 'Benjamin Štular, Edisa Lozić, Stefan Eichert'
__date__ = '2021-03-10'
__copyright__ = '(C) 2021 by Benjamin Štular, Edisa Lozić, Stefan Eichert'

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = '$Format:%H$'

"""
Model exported as python.
Name : Lidar Pipeline
Group : OpenLidarToolbox
With QGIS : 31604
"""

import inspect
import os
import pathlib
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtCore import QCoreApplication
from qgis.core import QgsProcessing
from qgis.core import QgsProcessingUtils
from qgis.core import QgsProcessingAlgorithm
from qgis.core import QgsProcessingMultiStepFeedback
from qgis.core import QgsProcessingParameterFile
from qgis.core import QgsProcessingParameterFileDestination
from qgis.core import QgsProcessingParameterNumber
from qgis.core import QgsProcessingParameterCrs
from qgis.core import QgsProcessingParameterBoolean
from qgis.core import QgsProcessingParameterString
from qgis.core import QgsProcessingParameterDefinition
import processing
from os.path import exists


class LidarPipeline(QgsProcessingAlgorithm):

    def initAlgorithm(self, config=None):

        self.addParameter(
            QgsProcessingParameterFile('InputFilelaslaz', 'Input File', behavior=QgsProcessingParameterFile.File,
                                       fileFilter='Lidar Files (*.las *.laz)', defaultValue=None))
        self.addParameter(QgsProcessingParameterCrs('CRS', 'Source File Coordinate System', defaultValue=None))
        self.addParameter(
            QgsProcessingParameterFileDestination('LAS', 'Classified LAZ', fileFilter='Lidar Files (*.laz)',
                                                  defaultValue=None, optional=False, createByDefault=True))
        self.addParameter(
            QgsProcessingParameterNumber('SetCellSize', 'Cell Size', type=QgsProcessingParameterNumber.Double,
                                         minValue=0, maxValue=1.79769e+308, defaultValue=0.5))
        self.addParameter(QgsProcessingParameterString('prefix', 'Name prefix for layers', multiLine=False,
                                                       defaultValue='ONE ', optional=False))
        self.addParameter(QgsProcessingParameterBoolean('VisualisationDFM', 'DFM', optional=False, defaultValue=True))
        self.addParameter(QgsProcessingParameterBoolean('GPD', 'Ground Point Density', optional=False, defaultValue=True))
        self.addParameter(QgsProcessingParameterBoolean('LVD', 'Low Vegetation Density', optional=False, defaultValue=True))
        self.addParameter(
            QgsProcessingParameterBoolean('VisualisationCM', 'DFM CM 0.5m', optional=False, defaultValue=True))
        self.addParameter(
            QgsProcessingParameterBoolean('VisualisationVAT', 'Visualisation VAT', optional=False, defaultValue=True))
        self.addParameter(
            QgsProcessingParameterBoolean('VisualisationSVF', 'Visualisation SVF', optional=False, defaultValue=True))
        self.addParameter(QgsProcessingParameterBoolean('VisualisationOPN', 'Visualisation Openness', optional=False,
                                                        defaultValue=True))
        self.addParameter(
            QgsProcessingParameterBoolean('VisualisationDfME', 'Visualisation DME', optional=False, defaultValue=True))
        self.addParameter(QgsProcessingParameterBoolean('VisualisationHS', 'Visualisation Hillshade', optional=False,
                                                        defaultValue=False))

    def processAlgorithm(self, parameters, context, model_feedback):
        # Use a multi-step feedback, so that individual child algorithm progress reports are adjusted for the
        # overall progress through the model
        feedback = QgsProcessingMultiStepFeedback(54, model_feedback)
        results = {}
        outputs = {}

        feedback.setCurrentStep(1)
        if feedback.isCanceled():
            return {}


        # lasground1
        alg_params = {
            'ADDITIONAL_OPTIONS': '',
            'BY_FLIGHTLINE': False,
            'CPU64': False,
            'GRANULARITY': 4,
            'GUI': False,
            'HORIZONTAL_FEET': False,
            'IGNORE_CLASS1': 8,
            'INPUT_LASLAZ': parameters['InputFilelaslaz'],
            'NO_BULGE': False,
            'OUTPUT_LASLAZ': QgsProcessingUtils.generateTempFilename('lasground1.laz'),
            'TERRAIN': 4,
            'VERBOSE': True,
            'VERTICAL_FEET': False
        }
        outputs['Lasground1'] = processing.run('LAStools:lasground', alg_params, context=context, feedback=feedback,
                                               is_child_algorithm=True)
        lasground1file = alg_params['OUTPUT_LASLAZ']

        feedback.setCurrentStep(1)
        if feedback.isCanceled():
            return {}

        # lasheight
        alg_params = {
            'ADDITIONAL_OPTIONS': '-ignore_class 0 1 3 4 5 6 7 8 9 10 11 12',
            'CPU64': False,
            'DROP_ABOVE': False,
            'DROP_ABOVE_HEIGHT': 100,
            'DROP_BELOW': False,
            'DROP_BELOW_HEIGHT': -2,
            'GUI': False,
            'IGNORE_CLASS1': 0,
            'IGNORE_CLASS2': 0,
            'INPUT_LASLAZ': lasground1file,
            'OUTPUT_LASLAZ': QgsProcessingUtils.generateTempFilename('lasheight.laz'),
            'REPLACE_Z': False,
            'VERBOSE': True
        }

        lasheightfile = alg_params['OUTPUT_LASLAZ']
        outputs['Lasheight'] = processing.run('LAStools:lasheight', alg_params, context=context, feedback=feedback,
                                              is_child_algorithm=True)

        feedback.setCurrentStep(2)
        if feedback.isCanceled():
            return {}

        # lasclassify
        alg_params = {
            'ADDITIONAL_OPTIONS': '-small_trees',
            'CPU64': False,
            'GUI': False,
            'HORIZONTAL_FEET': False,
            'IGNORE_CLASS1': 0,
            'IGNORE_CLASS2': 0,
            'INPUT_LASLAZ': lasheightfile,
            'OUTPUT_LASLAZ': QgsProcessingUtils.generateTempFilename('lasclassify.laz'),
            'VERBOSE': True,
            'VERTICAL_FEET': False
        }
        lasclassifyfile = alg_params['OUTPUT_LASLAZ']
        outputs['Lasclassify'] = processing.run('LAStools:lasclassify', alg_params, context=context, feedback=feedback,
                                                is_child_algorithm=True)

        feedback.setCurrentStep(3)
        if feedback.isCanceled():
            return {}

        # lasground2
        alg_params = {
            'ADDITIONAL_OPTIONS': '-ignore_class 6 7',
            'BY_FLIGHTLINE': False,
            'CPU64': False,
            'GRANULARITY': 4,
            'GUI': False,
            'HORIZONTAL_FEET': False,
            'IGNORE_CLASS1': 0,
            'INPUT_LASLAZ': lasclassifyfile,
            'NO_BULGE': False,
            'OUTPUT_LASLAZ': QgsProcessingUtils.generateTempFilename('lasground2.laz'),
            'TERRAIN': 1,
            'VERBOSE': True,
            'VERTICAL_FEET': False
        }
        lasground2file = alg_params['OUTPUT_LASLAZ']
        outputs['Lasground2'] = processing.run('LAStools:lasground', alg_params, context=context, feedback=feedback,
                                               is_child_algorithm=True)

        feedback.setCurrentStep(4)
        if feedback.isCanceled():
            return {}

        # lasheight_classify
        alg_params = {
            'ADDITIONAL_OPTIONS': '',
            'CLASSIFY_ABOVE': 6,
            'CLASSIFY_ABOVE_HEIGHT': 2,
            'CLASSIFY_BELOW': 8,
            'CLASSIFY_BELOW_HEIGHT': -0.25,
            'CLASSIFY_BETWEEN1': 3,
            'CLASSIFY_BETWEEN1_HEIGHT_FROM': -0.2,
            'CLASSIFY_BETWEEN1_HEIGHT_TO': 0.2,
            'CLASSIFY_BETWEEN2': 4,
            'CLASSIFY_BETWEEN2_HEIGHT_FROM': 0.5,
            'CLASSIFY_BETWEEN2_HEIGHT_TO': 2,
            'CPU64': False,
            'GUI': False,
            'IGNORE_CLASS1': 7,
            'IGNORE_CLASS2': 8,
            'INPUT_LASLAZ': lasground2file,
            'OUTPUT_LASLAZ': QgsProcessingUtils.generateTempFilename('lasheightclassify.las'),
            'REPLACE_Z': False,
            'VERBOSE': True
        }
        lasheightclassifyfile = alg_params['OUTPUT_LASLAZ']
        outputs['Lasheight_classify'] = processing.run('LAStools:lasheight_classify', alg_params, context=context,
                                                       feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(5)
        if feedback.isCanceled():
            return {}

        if parameters['LAS'] != 'TEMPORARY_OUTPUT':
            # lasduplicate
            alg_params = {
                'ADDITIONAL_OPTIONS': '-unique_xyz',
                'CPU64': False,
                'GUI': False,
                'HIGHEST_Z': False,
                'INPUT_LASLAZ': lasheightclassifyfile,
                'LOWEST_Z': False,
                'NEARBY': False,
                'NEARBY_TOLERANCE': 0.02,
                'OUTPUT_LASLAZ': parameters['LAS'],
                'RECORD_REMOVED': False,
                'SINGLE_RETURNS': False,
                'UNIQUE_XYZ': False,
                'VERBOSE': False
            }
            outputs['Lasduplicate'] = processing.run('LAStools:lasduplicate', alg_params, context=context,
                                                     feedback=feedback, is_child_algorithm=True)


        feedback.setCurrentStep(6)
        if feedback.isCanceled():
            return {}

        # LidarPointDensity1_Ground
        alg_params = {
            'exclude_cls': '0,1,3,4,5,7,8',
            'input': lasheightclassifyfile,
            'maxz': None,
            'minz': None,
            'radius': 10,
            'resolution': 2,
            'returns': 0,
            'output': QgsProcessingUtils.generateTempFilename('gpd.tif')
        }
        GroundPointDensityfile = alg_params['output']
        outputs['Lidarpointdensity1_ground'] = processing.run('wbt:LidarPointDensity', alg_params, context=context,
                                                              feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(7)
        if feedback.isCanceled():
            return {}

        # resampleGPD
        alg_params = {
            'GRASS_RASTER_FORMAT_META': '',
            'GRASS_RASTER_FORMAT_OPT': '',
            'GRASS_REGION_CELLSIZE_PARAMETER': parameters['SetCellSize'],
            'GRASS_REGION_PARAMETER': GroundPointDensityfile,
            'input': GroundPointDensityfile,
            'output': QgsProcessingUtils.generateTempFilename('GPDres.tif')
        }
        GPDfileR = alg_params['output']
        outputs['Resamplegpd'] = processing.run('grass7:r.resample', alg_params, context=context, feedback=feedback,
                                                is_child_algorithm=True)

        feedback.setCurrentStep(8)
        if feedback.isCanceled():
            return {}

        # Warp (reproject)
        alg_params = {
            'DATA_TYPE': 0,
            'EXTRA': '',
            'INPUT': GPDfileR,
            'MULTITHREADING': False,
            'NODATA': None,
            'OPTIONS': '',
            'RESAMPLING': 0,
            'SOURCE_CRS': parameters['CRS'],
            'TARGET_CRS': parameters['CRS'],
            'TARGET_EXTENT': None,
            'TARGET_EXTENT_CRS': None,
            'TARGET_RESOLUTION': None,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('GPDres.tif')
        }

        outputs['WarpReproject'] = processing.run('gdal:warpreproject', alg_params, context=context,
                                                  feedback=feedback,
                                                  is_child_algorithm=True)
        GPDfileR = alg_params['OUTPUT']

        feedback.setCurrentStep(9)
        if feedback.isCanceled():
            return {}

        # Load layer into project
        if parameters['GPD']:
            alg_params = {
                'INPUT': GPDfileR,
                'NAME': parameters['prefix'] + 'Ground Point Density'
            }
            outputs['LoadLayerIntoProject'] = processing.run('native:loadlayer', alg_params, context=context,
                                                             feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(10)
        if feedback.isCanceled():
            return {}

        # LidarPointDensity2_LowVeg
        alg_params = {
            'exclude_cls': '0,1,2,4,5,6,7',
            'input': lasheightclassifyfile,
            'maxz': None,
            'minz': None,
            'radius': 10,
            'resolution': 2,
            'returns': 0,
            'output': QgsProcessingUtils.generateTempFilename('lvd.tif')
        }
        LowVegfile = alg_params['output']
        outputs['Lidarpointdensity2_lowveg'] = processing.run('wbt:LidarPointDensity', alg_params, context=context,
                                                              feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(11)
        if feedback.isCanceled():
            return {}

        # resampleLVD
        alg_params = {
            'GRASS_RASTER_FORMAT_META': '',
            'GRASS_RASTER_FORMAT_OPT': '',
            'GRASS_REGION_CELLSIZE_PARAMETER': parameters['SetCellSize'],
            'GRASS_REGION_PARAMETER': GroundPointDensityfile,
            'input': LowVegfile,
            'output': QgsProcessingUtils.generateTempFilename('LVD.tif')
        }
        LVDfileR = alg_params['output']
        outputs['Resamplelvd'] = processing.run('grass7:r.resample', alg_params, context=context, feedback=feedback,
                                                is_child_algorithm=True)

        feedback.setCurrentStep(12)
        if feedback.isCanceled():
            return {}

        # Warp (reproject)
        alg_params = {
            'DATA_TYPE': 0,
            'EXTRA': '',
            'INPUT': LVDfileR,
            'MULTITHREADING': False,
            'NODATA': None,
            'OPTIONS': '',
            'RESAMPLING': 0,
            'SOURCE_CRS': parameters['CRS'],
            'TARGET_CRS': parameters['CRS'],
            'TARGET_EXTENT': None,
            'TARGET_EXTENT_CRS': None,
            'TARGET_RESOLUTION': None,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('LVD.tif')
        }

        outputs['WarpReproject'] = processing.run('gdal:warpreproject', alg_params, context=context,
                                                  feedback=feedback,
                                                  is_child_algorithm=True)
        LVDfileR = alg_params['OUTPUT']

        feedback.setCurrentStep(13)
        if feedback.isCanceled():
            return {}

        # Load layer into project
        if parameters['LVD']:
            alg_params = {
                'INPUT': LVDfileR,
                'NAME': parameters['prefix'] + 'Low Vegetation Density'
            }
            outputs['LoadLayerIntoProject'] = processing.run('native:loadlayer', alg_params, context=context,
                                                             feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(14)
        if feedback.isCanceled():
            return {}

        # LidarTINGridding
        alg_params = {
            'exclude_cls': '0,1,3,4,5,7',
            'input': lasheightclassifyfile,
            'max_triangle_edge_length': None,
            'maxz': None,
            'minz': None,
            'parameter': 0,
            'resolution': 0.5,
            'returns': 0,
            'output': QgsProcessingUtils.generateTempFilename('tin.tif')
        }
        tingrid = alg_params['output']
        outputs['Lidartingridding'] = processing.run('wbt:LidarTINGridding', alg_params, context=context,
                                                     feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(15)
        if feedback.isCanceled():
            return {}

        # resampleTIN
        alg_params = {
            'GRASS_RASTER_FORMAT_META': '',
            'GRASS_RASTER_FORMAT_OPT': '',
            'GRASS_REGION_CELLSIZE_PARAMETER': parameters['SetCellSize'],
            'GRASS_REGION_PARAMETER': GroundPointDensityfile,
            'input': tingrid,
            'output': QgsProcessingUtils.generateTempFilename('tinRes.tif')
        }
        TINfileR = alg_params['output']
        outputs['ResampleTin'] = processing.run('grass7:r.resample', alg_params, context=context, feedback=feedback,
                                                is_child_algorithm=True)

        feedback.setCurrentStep(16)
        if feedback.isCanceled():
            return {}

        # Warp (reproject)
        alg_params = {
            'DATA_TYPE': 0,
            'EXTRA': '',
            'INPUT': TINfileR,
            'MULTITHREADING': False,
            'NODATA': None,
            'OPTIONS': '',
            'RESAMPLING': 0,
            'SOURCE_CRS': parameters['CRS'],
            'TARGET_CRS': parameters['CRS'],
            'TARGET_EXTENT': None,
            'TARGET_EXTENT_CRS': None,
            'TARGET_RESOLUTION': None,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('TINfileR.tif')
        }

        outputs['WarpReproject'] = processing.run('gdal:warpreproject', alg_params, context=context, feedback=feedback,
                                                  is_child_algorithm=True)
        TINfileR = alg_params['OUTPUT']

        feedback.setCurrentStep(17)
        if feedback.isCanceled():
            return {}

        # LidarIdwInterpolation
        alg_params = {
            'exclude_cls': '0,1,3,4,5,7,8,9',
            'input': lasheightclassifyfile,
            'maxz': None,
            'minz': None,
            'parameter': 0,
            'radius': 5,
            'resolution': 0.5,
            'returns': 0,
            'weight': 2,
            'output': QgsProcessingUtils.generateTempFilename('idw.tif')
        }
        lidarIDW = alg_params['output']
        outputs['Lidaridwinterpolation'] = processing.run('wbt:LidarIdwInterpolation', alg_params, context=context,
                                                          feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(18)
        if feedback.isCanceled():
            return {}

        # resampleIDW
        alg_params = {
            'GRASS_RASTER_FORMAT_META': '',
            'GRASS_RASTER_FORMAT_OPT': '',
            'GRASS_REGION_CELLSIZE_PARAMETER': parameters['SetCellSize'],
            'GRASS_REGION_PARAMETER': GroundPointDensityfile,
            'input': lidarIDW,
            'output': QgsProcessingUtils.generateTempFilename('IDWres.tif')
        }
        IDWfileR = alg_params['output']
        outputs['ResampleIDW'] = processing.run('grass7:r.resample', alg_params, context=context, feedback=feedback,
                                                is_child_algorithm=True)

        feedback.setCurrentStep(19)
        if feedback.isCanceled():
            return {}

        # Warp (reproject)
        alg_params = {
            'DATA_TYPE': 0,
            'EXTRA': '',
            'INPUT': IDWfileR,
            'MULTITHREADING': False,
            'NODATA': None,
            'OPTIONS': '',
            'RESAMPLING': 0,
            'SOURCE_CRS': parameters['CRS'],
            'TARGET_CRS': parameters['CRS'],
            'TARGET_EXTENT': None,
            'TARGET_EXTENT_CRS': None,
            'TARGET_RESOLUTION': None,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('IDWres.tif')
        }

        outputs['WarpReproject'] = processing.run('gdal:warpreproject', alg_params, context=context, feedback=feedback,
                                                  is_child_algorithm=True)
        IDWfileR = alg_params['OUTPUT']

        feedback.setCurrentStep(20)
        if feedback.isCanceled():
            return {}

        # Confidence Map 0.5m
        # Density Hi
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': GPDfileR,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 2,
            'RASTER_BAND': 1,
            'TABLE': [0, 4, 0, 4.000000001, 100000, 1],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('densHi.tif')
        }
        densHi = alg_params['OUTPUT']
        outputs['DensityHi'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                              feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(21)
        if feedback.isCanceled():
            return {}

        # Density Mid
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': GPDfileR,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 2,
            'RASTER_BAND': 1,
            'TABLE': [0, 2, 0, 2.000001, 4, 1, 4.000001, 100000, 0],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('densMid.tif')
        }
        densMid = alg_params['OUTPUT']
        outputs['DensityMid'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                               feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(22)
        if feedback.isCanceled():
            return {}

        # Density Low
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': GPDfileR,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 1,
            'RASTER_BAND': 1,
            'TABLE': [0, 1, 0, 1.00000001, 2, 1, 2.00000001, 100000, 0],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('densLow.tif')
        }
        densLow = alg_params['OUTPUT']
        outputs['DensityLow'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                               feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(23)
        if feedback.isCanceled():
            return {}

        # Density Vlow
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': GPDfileR,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 1,
            'RASTER_BAND': 1,
            'TABLE': [0, 1, 1, 1.00000001, 100000, 0],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('densVlow.tif')
        }
        densVlow = alg_params['OUTPUT']
        outputs['DensityVlow'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                                feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(24)
        if feedback.isCanceled():
            return {}

        # VegLow
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': LVDfileR,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 2,
            'RASTER_BAND': 1,
            'TABLE': [0, 4, 1, 4.00000001, 100000, 0],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('veglow.tif')
        }
        veglow = alg_params['OUTPUT']
        outputs['Veglow'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                           feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(25)
        if feedback.isCanceled():
            return {}

        # VegHi
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': LVDfileR,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 2,
            'RASTER_BAND': 1,
            'TABLE': [0, 4, 0, 4.00000001, 100000, 1],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('veghi.tif')
        }
        veghi = alg_params['OUTPUT']
        outputs['Veghi'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                          feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(26)
        if feedback.isCanceled():
            return {}

        # Slope Qgis
        alg_params = {
            'INPUT': TINfileR,
            'Z_FACTOR': 1,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('slope.tif')
        }
        sloperast = alg_params['OUTPUT']
        outputs['SlopeQgis'] = processing.run('native:slope', alg_params, context=context, feedback=feedback,
                                              is_child_algorithm=True)

        feedback.setCurrentStep(27)
        if feedback.isCanceled():
            return {}

        # resampleSlope
        alg_params = {
            'GRASS_RASTER_FORMAT_META': '',
            'GRASS_RASTER_FORMAT_OPT': '',
            'GRASS_REGION_CELLSIZE_PARAMETER': parameters['SetCellSize'],
            'GRASS_REGION_PARAMETER': GroundPointDensityfile,
            'input': sloperast,
            'output': QgsProcessingUtils.generateTempFilename('SlopeRes.tif')
        }
        sloperaster = alg_params['output']
        outputs['ResampleSlope'] = processing.run('grass7:r.resample', alg_params, context=context, feedback=feedback,
                                                  is_child_algorithm=True)

        feedback.setCurrentStep(28)
        if feedback.isCanceled():
            return {}

        # Slope 12
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': sloperaster,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 2,
            'RASTER_BAND': 1,
            'TABLE': [0, 12.5, 1, 12.5000001, 90, 0],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('slope12.tif')
        }
        slope12 = alg_params['OUTPUT']
        outputs['Slope12'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                            feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(29)
        if feedback.isCanceled():
            return {}

        # Slope 22
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': sloperaster,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 2,
            'RASTER_BAND': 1,
            'TABLE': [0, 12.5, 0, 12.500000001, 22.5, 1, 22.500000001, 90, 0],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('slope22.tif')
        }
        slope22 = alg_params['OUTPUT']
        outputs['Slope22'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                            feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(30)
        if feedback.isCanceled():
            return {}

        # Slope 42
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': sloperaster,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 2,
            'RASTER_BAND': 1,
            'TABLE': [0, 22.5, 0, 22.50000001, 42.5, 1, 42.50000001, 90, 0],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('slope42.tif')
        }
        slope42 = alg_params['OUTPUT']
        outputs['Slope42'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                            feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(31)
        if feedback.isCanceled():
            return {}

        # Slope 90
        alg_params = {
            'DATA_TYPE': 3,
            'INPUT_RASTER': sloperaster,
            'NODATA_FOR_MISSING': True,
            'NO_DATA': -9999,
            'RANGE_BOUNDARIES': 2,
            'RASTER_BAND': 1,
            'TABLE': [0, 42.5, 0, 42.50000001, 90, 1],
            'OUTPUT': QgsProcessingUtils.generateTempFilename('slope90.tif')
        }
        slope90 = alg_params['OUTPUT']
        outputs['Slope90'] = processing.run('native:reclassifybytable', alg_params, context=context,
                                            feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(32)
        if feedback.isCanceled():
            return {}

        # Calc 1a
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': None,
            'BAND_E': None,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': '(A+B+C)*D',
            'INPUT_A': densMid,
            'INPUT_B': densLow,
            'INPUT_C': densVlow,
            'INPUT_D': slope90,
            'INPUT_E': None,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('calc1a.tif')
        }
        calc1a = alg_params['OUTPUT']
        outputs['Calc1a'] = processing.run('gdal:rastercalculator', alg_params, context=context, feedback=feedback,
                                           is_child_algorithm=True)

        feedback.setCurrentStep(33)
        if feedback.isCanceled():
            return {}

        # Calc 1b
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': None,
            'BAND_E': None,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': 'A*(B+C)',
            'INPUT_A': densVlow,
            'INPUT_B': slope12,
            'INPUT_C': slope22,
            'INPUT_D': None,
            'INPUT_E': None,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('calc1b.tif')
        }
        calc1b = alg_params['OUTPUT']
        outputs['Calc1b'] = processing.run('gdal:rastercalculator', alg_params, context=context, feedback=feedback,
                                           is_child_algorithm=True)

        feedback.setCurrentStep(34)
        if feedback.isCanceled():
            return {}

        # Calc 2
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': None,
            'BAND_E': None,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': '(A+B+C)*D*2',
            'INPUT_A': densMid,
            'INPUT_B': densLow,
            'INPUT_C': densVlow,
            'INPUT_D': slope42,
            'INPUT_E': None,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('calc2.tif')
        }
        calc2 = alg_params['OUTPUT']
        outputs['Calc2'] = processing.run('gdal:rastercalculator', alg_params, context=context, feedback=feedback,
                                          is_child_algorithm=True)

        feedback.setCurrentStep(35)
        if feedback.isCanceled():
            return {}

        # Calc 3
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': None,
            'BAND_E': None,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': 'A*(B+C)*3',
            'INPUT_A': densLow,
            'INPUT_B': slope12,
            'INPUT_C': slope22,
            'INPUT_D': None,
            'INPUT_E': None,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('calc3.tif')
        }
        calc3 = alg_params['OUTPUT']
        outputs['Calc3'] = processing.run('gdal:rastercalculator', alg_params, context=context, feedback=feedback,
                                          is_child_algorithm=True)

        feedback.setCurrentStep(36)
        if feedback.isCanceled():
            return {}

        # Calc 4a
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': 1,
            'BAND_E': 1,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': 'A*(B+C+D)*E*4',
            'INPUT_A': densHi,
            'INPUT_B': slope22,
            'INPUT_C': slope42,
            'INPUT_D': slope90,
            'INPUT_E': veghi,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('calc4a.tif')
        }
        calc4a = alg_params['OUTPUT']
        outputs['Calc4a'] = processing.run('gdal:rastercalculator', alg_params, context=context, feedback=feedback,
                                           is_child_algorithm=True)

        feedback.setCurrentStep(37)
        if feedback.isCanceled():
            return {}

        # Calc 4b
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': None,
            'BAND_E': None,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': 'A*(B+C)*4',
            'INPUT_A': densMid,
            'INPUT_B': slope12,
            'INPUT_C': slope22,
            'INPUT_D': None,
            'INPUT_E': None,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('Calc4b.tif')
        }
        calc4b = alg_params['OUTPUT']
        outputs['Calc4b'] = processing.run('gdal:rastercalculator', alg_params, context=context, feedback=feedback,
                                           is_child_algorithm=True)

        feedback.setCurrentStep(38)
        if feedback.isCanceled():
            return {}

        # Calc 5a
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': None,
            'BAND_E': None,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': 'A*B*C*5',
            'INPUT_A': densHi,
            'INPUT_B': slope12,
            'INPUT_C': veghi,
            'INPUT_D': None,
            'INPUT_E': None,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('calc5a.tif')
        }
        calc5a = alg_params['OUTPUT']
        outputs['Calc5a'] = processing.run('gdal:rastercalculator', alg_params, context=context, feedback=feedback,
                                           is_child_algorithm=True)

        feedback.setCurrentStep(39)
        if feedback.isCanceled():
            return {}

        # Calc 5b
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': 1,
            'BAND_E': 1,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': 'A*(B+C+D)*E*5',
            'INPUT_A': densHi,
            'INPUT_B': slope22,
            'INPUT_C': slope42,
            'INPUT_D': slope90,
            'INPUT_E': veglow,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('calc5b.tif')
        }
        calc5b = alg_params['OUTPUT']
        outputs['Calc5b'] = processing.run('gdal:rastercalculator', alg_params, context=context, feedback=feedback,
                                           is_child_algorithm=True)

        feedback.setCurrentStep(40)
        if feedback.isCanceled():
            return {}

        # Calc 6
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': None,
            'BAND_E': None,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': 'A*B*C*6',
            'INPUT_A': densHi,
            'INPUT_B': slope12,
            'INPUT_C': veglow,
            'INPUT_D': None,
            'INPUT_E': None,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('calc6.tif')
        }
        calc6 = alg_params['OUTPUT']
        outputs['Calc6'] = processing.run('gdal:rastercalculator', alg_params, context=context, feedback=feedback,
                                          is_child_algorithm=True)

        feedback.setCurrentStep(41)
        if feedback.isCanceled():
            return {}

        # CalcCran1
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': 1,
            'BAND_E': 1,
            'BAND_F': 1,
            'EXTRA': '',
            'FORMULA': 'A+B+C+D+E+F',
            'INPUT_A': calc6,
            'INPUT_B': calc5a,
            'INPUT_C': calc5b,
            'INPUT_D': calc4a,
            'INPUT_E': calc4b,
            'INPUT_F': calc3,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('calccran1.tif')
        }
        calccran1 = alg_params['OUTPUT']
        outputs['Calccran1'] = processing.run('gdal:rastercalculator', alg_params, context=context,
                                              feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(42)
        if feedback.isCanceled():
            return {}

        # CalcCranFinal
        alg_params = {
            'BAND_A': 1,
            'BAND_B': 1,
            'BAND_C': 1,
            'BAND_D': 1,
            'BAND_E': None,
            'BAND_F': None,
            'EXTRA': '',
            'FORMULA': 'A+B+C+D',
            'INPUT_A': calccran1,
            'INPUT_B': calc2,
            'INPUT_C': calc1a,
            'INPUT_D': calc1b,
            'INPUT_E': None,
            'INPUT_F': None,
            'NO_DATA': None,
            'OPTIONS': '',
            'RTYPE': 4,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('CFM.tif')
        }
        cfm = alg_params['OUTPUT']
        outputs['Calccranfinal'] = processing.run('gdal:rastercalculator', alg_params, context=context,
                                                  feedback=feedback, is_child_algorithm=True)
        results['CFM'] = outputs['Calccranfinal']['OUTPUT']

        feedback.setCurrentStep(43)
        if feedback.isCanceled():
            return {}

        # Warp (reproject)
        alg_params = {
            'DATA_TYPE': 0,
            'EXTRA': '',
            'INPUT': cfm,
            'MULTITHREADING': False,
            'NODATA': None,
            'OPTIONS': '',
            'RESAMPLING': 0,
            'SOURCE_CRS': parameters['CRS'],
            'TARGET_CRS': parameters['CRS'],
            'TARGET_EXTENT': None,
            'TARGET_EXTENT_CRS': None,
            'TARGET_RESOLUTION': None,
            'OUTPUT': QgsProcessingUtils.generateTempFilename('cfm.tif')
        }

        outputs['WarpReproject'] = processing.run('gdal:warpreproject', alg_params, context=context,
                                                  feedback=feedback,
                                                  is_child_algorithm=True)
        cfm = alg_params['OUTPUT']

        feedback.setCurrentStep(44)
        if feedback.isCanceled():
            return {}

        # Set style for raster layer
        folder = os.path.split(inspect.getfile(inspect.currentframe()))[0]
        styleFile = os.path.join(os.path.join(folder, 'DFMconfidenceMap.qml'))

        alg_params = {
            'INPUT': cfm,
            'STYLE': styleFile
        }
        outputs['SetStyleForRasterLayer'] = processing.run('qgis:setstyleforrasterlayer', alg_params,
                                                           context=context,
                                                           feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(45)
        if feedback.isCanceled():
            return {}

        # Load layer into project
        if parameters['VisualisationCM']:
            alg_params = {
                'INPUT': cfm,
                'NAME': parameters['prefix'] + 'DFM CM 0.5m'
            }
            outputs['LoadLayerIntoProject'] = processing.run('native:loadlayer', alg_params, context=context,
                                                             feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(46)
        if feedback.isCanceled():
            return {}

        # Hybrid Interpolation
        alg_params = {
            'CellSize': 0.5,
            'ConfidenceMapRaster': cfm,
            'REDgrowradiusinrastercells': 3,
            'TLI': TINfileR,
            'IDW': IDWfileR,
            'Dfm': QgsProcessingUtils.generateTempFilename('dfm1.tif'),
            'loadDFM': False
        }
        outputs['HybridInterpolation'] = processing.run('Open LiDAR Toolbox:Hybrid Interpolation', alg_params,
                                                        context=context, feedback=feedback, is_child_algorithm=True)

        DFMPatched = outputs['HybridInterpolation']['Dfm']

        feedback.setCurrentStep(47)
        if feedback.isCanceled():
            return {}

        if parameters['VisualisationDFM']:
            # Load layer into project
            alg_params = {
                'INPUT': DFMPatched,
                'NAME': parameters['prefix'] + 'DFM'
            }
            outputs['LoadLayerIntoProject'] = processing.run('native:loadlayer', alg_params, context=context,
                                                             feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(48)
        if feedback.isCanceled():
            return {}

        if parameters['VisualisationSVF']:
            # RVT Sky-view factor
            alg_params = {
                'FILL_NO_DATA': False,
                'FILL_METHOD': 0,
                'INPUT': DFMPatched,
                'KEEP_ORIG_NO_DATA': False,
                'NOISE_REMOVE': 0,
                'NUM_DIRECTIONS': 16,
                'RADIUS': 10,
                'SAVE_AS_8BIT': False,
                'VE_FACTOR': 1,
                'OUTPUT': QgsProcessingUtils.generateTempFilename('SKF.tif')
            }
            outputs['RvtSkyviewFactor'] = processing.run('rvt:rvt_svf', alg_params, context=context, feedback=feedback,
                                                         is_child_algorithm=True)

            # Load layer into project
            alg_params = {
                'INPUT': outputs['RvtSkyviewFactor']['OUTPUT'],
                'NAME': parameters['prefix'] + 'Visualisation SVF'
            }
            outputs['LoadLayerIntoProject'] = processing.run('native:loadlayer', alg_params, context=context,
                                                             feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(49)
        if feedback.isCanceled():
            return {}

        if parameters['VisualisationVAT']:
            # RVT Blender
            alg_params = {
                'BLEND_COMBINATION': 0,
                'FILL_NO_DATA': False,
                'FILL_METHOD': 0,
                'INPUT': DFMPatched,
                'KEEP_ORIG_NO_DATA': False,
                'OUTPUT': QgsProcessingUtils.generateTempFilename('vat.tif'),
                'TERRAIN_TYPE': 0
            }
            outputs['RvtBlender'] = processing.run('rvt:rvt_blender', alg_params, context=context, feedback=feedback,
                                                   is_child_algorithm=True)

            feedback.setCurrentStep(20)
            if feedback.isCanceled():
                return {}

            # Load layer into project
            alg_params = {
                'INPUT': outputs['RvtBlender']['OUTPUT'],
                'NAME': parameters['prefix'] + 'Visualisation VAT'
            }
            outputs['LoadLayerIntoProject'] = processing.run('native:loadlayer', alg_params, context=context,
                                                             feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(50)
        if feedback.isCanceled():
            return {}

        if parameters['VisualisationOPN']:
            # RVT Openness
            alg_params = {
                'FILL_NO_DATA': False,
                'FILL_METHOD': 0,
                'INPUT': DFMPatched,
                'KEEP_ORIG_NO_DATA': False,
                'NOISE_REMOVE': 0,
                'NUM_DIRECTIONS': 16,
                'OPNS_TYPE': 0,
                'RADIUS': 10,
                'SAVE_AS_8BIT': False,
                'VE_FACTOR': 1,
                'OUTPUT': QgsProcessingUtils.generateTempFilename('OPN.tif')
            }
            outputs['RvtOpenness'] = processing.run('rvt:rvt_opns', alg_params, context=context, feedback=feedback,
                                                    is_child_algorithm=True)



            # Load layer into project
            alg_params = {
                'INPUT': outputs['RvtOpenness']['OUTPUT'],
                'NAME': parameters['prefix'] + 'Visualisation Openness'
            }
            outputs['LoadLayerIntoProject'] = processing.run('native:loadlayer', alg_params, context=context,
                                                             feedback=feedback, is_child_algorithm=True)

        feedback.setCurrentStep(51)
        if feedback.isCanceled():
            return {}

        if parameters['VisualisationDfME']:
            # DiffFromMeanElev
            alg_params = {
                'filterx': 10,
                'filtery': 10,
                'input': DFMPatched,
                'output': QgsProcessingUtils.generateTempFilename('DME.tif')
            }
            outputs['Difffrommeanelev'] = processing.run('wbt:DiffFromMeanElev', alg_params, context=context,
                                                         feedback=feedback, is_child_algorithm=True)

            # Set style for raster layer
            folder = os.path.split(inspect.getfile(inspect.currentframe()))[0]
            styleFile = os.path.join(os.path.join(folder, 'DfME.qml'))

            alg_params = {
                'INPUT': outputs['Difffrommeanelev']['output'],
                'STYLE': styleFile
            }
            outputs['SetStyleForRasterLayer'] = processing.run('qgis:setstyleforrasterlayer', alg_params,
                                                               context=context,
                                                               feedback=feedback, is_child_algorithm=True)

            # Load layer into project
            alg_params = {
                'INPUT': outputs['Difffrommeanelev']['output'],
                'NAME': parameters['prefix'] + 'Visualisation DME'
            }
            outputs['LoadLayerIntoProject'] = processing.run('native:loadlayer', alg_params, context=context,
                                                             feedback=feedback, is_child_algorithm=True)
        feedback.setCurrentStep(52)
        if feedback.isCanceled():
            return {}

        if parameters['VisualisationHS']:
            # Hillshade
            alg_params = {
                'AZIMUTH': 300,
                'INPUT': DFMPatched,
                'V_ANGLE': 40,
                'Z_FACTOR': 1,
                'OUTPUT': QgsProcessing.TEMPORARY_OUTPUT
            }
            outputs['Hillshade'] = processing.run('native:hillshade', alg_params, context=context, feedback=feedback,
                                                  is_child_algorithm=True)


            # Load layer into project
            alg_params = {
                'INPUT': outputs['Hillshade']['OUTPUT'],
                'NAME': parameters['prefix'] + 'Visualisation Hillshade'
            }
            outputs['LoadLayerIntoProject'] = processing.run('native:loadlayer', alg_params, context=context,
                                                             feedback=feedback, is_child_algorithm=True)
        feedback.setCurrentStep(53)
        if feedback.isCanceled():
            return {}

        return results

    def name(self):
        return 'ONE'

    def displayName(self):
        return 'ONE (One-Step-Processing)'

    def group(self):
        """
        Returns the name of the group this algorithm belongs to. This string
        should be localised.
        """
        return self.tr(self.groupId())

    def icon(self):
        cmd_folder = os.path.split(inspect.getfile(inspect.currentframe()))[0]
        icon = QIcon(os.path.join(os.path.join(cmd_folder, 'one.png')))
        return icon

    def groupId(self):
        """
        Returns the unique ID of the group this algorithm belongs to. This
        string should be fixed for the algorithm, and must not be localised.
        The group id should be unique within each provider. Group id should
        contain lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return ''

    def shortHelpString(self):
        return """<html><body><h2>Algorithm description</h2>
    <p>This is an algorithm pipeline that takes an unclassified airborne LiDAR point cloud to produce all derivatives essential for archaeology and anyone interested in visual analysis of LiDAR data.</p>
    <h2>Input parameters</h2>
    <h3>Input File</h3>
    <p>Unclassified point cloud in LAS or LAZ format. Noise classified as ASPRS class 7 will be exempt from the processing, all other preexisting classification will be ignored.</p>
    <h3>Source File Coordinate System</h3>
    <p>Select the Coordinate Reference System (CRS) of the input LAS/LAZ file. Make sure that the CRS is Cartesian (x and y in meters, not degrees). If you are not sure which the is correct CRS and you only need it temporarily you can choose any Cartesian CRS, for example, EPSG:8687.</p>
    <h3>Cell Size</h3>
    <p>DFM grid resolution, default value is 0.5 m. Optimal resolution for any given point cloud can be calculated with the DFM Confidence Map tool.</p>
    <h3>Name prefix for layers</h3>
    <p>The output layers are added to the map as temporary layers with default names. They can be saved as files afterwards. In order to distinguish them from previously created files with the same tool a prefix should be defined to avoid the same names for different layers</p>
    <p><b>Classified LAZ: </b> Classified point cloud. QGIS cannot load point clouds so it must be saved as a LAZ file. Specify folder and file name.</p>
    <h3>Outputs:</h3>
    <p><b>DFM: </b> Digital feature model, which is a type of DEM that combines ground and buildings</p>
    <p><b>Ground Point Density</b></p>
    <p><b>Low Vegetation Density</b></p>
    <p><b>DFM CM 0.5m: </b> DFM Confidence Map for 0.5 m resolution (if other resolutions are needed – e.g., the map is either completely red or completely blue – use the dedicated tool)</p>
    <p><b>Classified LAZ: </b> Classified point cloud. This can saved to a ".laz" file which is not added to the map.</p>
    <p><b>Visualisations:</b></p>
    <p><b>VAT: </b> Visualisation for archaeological topography</p>
    <p><b>SVF: </b> Sky view factor</p>
    <p><b>Opennes: </b> Openness – positive</p>
    <p><b>DME: </b> Difference from mean elevation</p>
    <p><b>Hillshade: </b> Hillshade/Relief of DFM</p>
    <h2>FAQ</h2>
    <h3>The edges of my outputs are black</h3>
    <p>This is due to the so called edge effect. In many steps the values are calculated from surrounding points; since at the edge there are no surrounding points, the output values are "strange", e.g., showing as black on most visualisations. This cannot be avoided and the only solution is to process larger areas or to create overlapping mosaics.</p>
    <p></p>
    <br><br>
    <p><b>Literature:</b> Štular, Lozić, Eichert 2021 (in press).</p>
    <br><a href="https://github.com/stefaneichert/OpenLidarTools">Website</a>
    <br><p align="right">Algorithm author: Benjamin Štular, Edisa Lozić, Stefan Eichert </p><p align="right">Help author: Benjamin Štular, Edisa Lozić, Stefan Eichert</p></body></html>"""

    def tr(self, string):
        return QCoreApplication.translate('Processing', string)

    def createInstance(self):
        return LidarPipeline()
